/*
*
* Driver for H3600 Extension Packs
*
* Copyright 2000 Compaq Computer Corporation.
*
* Use consistent with the GNU GPL is permitted,
* provided that this copyright notice is
* preserved in its entirety in all copies and derived works.
*
* COMPAQ COMPUTER CORPORATION MAKES NO WARRANTIES, EXPRESSED OR IMPLIED,
* AS TO THE USEFULNESS OR CORRECTNESS OF THIS CODE OR ITS
* FITNESS FOR ANY PARTICULAR PURPOSE.
*
* Author: Jamey Hicks.
*
*/

#include <linux/module.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>

/* SA1100 serial defines */
#include <asm/arch/hardware.h>
#include <asm/arch/serial_reg.h>
#include <asm/arch/irqs.h>

#include <linux/h3600_ts.h>

static void msleep(unsigned int msec)
{
	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout( (msec * HZ + 999) / 1000);
}

static int h3600_sleeve_major = 0;

struct {
        char start_of_id; /* 0xaa */
        int data_len;
        char version;
        short vendor_id;
        short device_id;
} sleeve_header;

struct proc_dir_entry *proc_sleeve_dir = NULL;

static int h3600_sleeve_proc_read_eeprom(char *page, char **start, off_t off, int count, int *eof, void *data)
{
        char *p = page;
        int len = 0;
        int i;
        char mark = 0;
        int offset = 1;

        h3600_iic_read(0, &mark, 1);
        h3600_iic_read(1, &len, 4);
        if (mark != 0xaa || len == 0xFFFFFFFF)
                return 0;
        printk(__FUNCTION__ ": len=%d\n", len);
        for (i = 0; i < len; i+=8) {
                h3600_iic_read(offset, p, 8);
                p += 8;
                offset += 8;
        }
        if (len != (p - page))
                h3600_iic_read(offset, p, len-(p - page));
        return len;
}

static int h3600_sleeve_proc_read_backpaq(char *page, char **start, off_t off, int count, int *eof, void *data)
{
        char *p = page;
        p += sprintf(p, "SM2: %#08x\n", *(long *)0xf1000000);
        p += sprintf(p, "SM4: %#08x\n", *(long *)0xf2000000);
        p += sprintf(p, "PCMCIA SKT1 IO: %#08x\n", *(long *)0xf6000000);
        p += sprintf(p, "PCMCIA SKT2 IO: %#08x\n", *(long *)0xf7000000);
        return p-page;
}

DECLARE_WAIT_QUEUE_HEAD(h3600_sleeve_wait);
static int h3600_sleeve_active = 1;
static int h3600_sleeve_thread(void *x)
{
        struct task_struct *tsk = current;

        tsk->session = 1;
        tsk->pgrp = 1;
        strcpy(tsk->comm, "k_h3600_sleeve");
        sigfillset(&tsk->blocked);

        while (h3600_sleeve_active) {
                int opt_ndet;
                interruptible_sleep_on(&h3600_sleeve_wait);
                /* debounce */
                printk(__FUNCTION__ ": before first msleep\n"); 
                msleep(100);
                opt_ndet = (GPLR & GPIO_BITSY_OPT_DET);
                if (!opt_ndet) {
                        set_bitsy_egpio(EGPIO_BITSY_OPT_NVRAM_ON | EGPIO_BITSY_OPT_ON);
                } else {
                        clr_bitsy_egpio(EGPIO_BITSY_OPT_NVRAM_ON | EGPIO_BITSY_OPT_ON);
                }
                msleep(100);
                printk(__FUNCTION__ ": x=%p opt_det=%#x\n", x, !opt_ndet);

        }
        return 0;
}

static void h3600_sleeve_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
        int opt_ndet = (GPLR & GPIO_BITSY_OPT_DET);
        printk(__FUNCTION__ ": OPT_DET=%#08x\n", !opt_ndet);
        /* now clear the interrupt */
        GEDR = GPIO_BITSY_OPT_DET;
        wake_up_interruptible(&h3600_sleeve_wait);
}




int __init h3600_sleeve_init_module(void)
{
	int result;

        if (!machine_is_bitsy())
          return -ENODEV;

        /* register our character device */
#if 0
        printk(__FUNCTION__ ": registering char device\n");
        result = register_chrdev(0, MOD_NAME, &ts_fops);
        if (result < 0) {
                printk(__FUNCTION__ ": can't get major number\n");
                return result;
        }

	h3600_sleeve_major = result;
#endif

        clr_bitsy_egpio(EGPIO_BITSY_OPT_NVRAM_ON | EGPIO_BITSY_OPT_ON);

        set_GPIO_IRQ_edge( GPIO_BITSY_OPT_DET, GPIO_BOTH_EDGES );
	result = request_irq(IRQ_GPIO_BITSY_OPT_DET, h3600_sleeve_interrupt, SA_SHIRQ | SA_INTERRUPT | SA_SAMPLE_RANDOM,
                             "h3600_sleeve", h3600_sleeve_interrupt);

        proc_sleeve_dir = proc_mkdir("sleeve", proc_bus);
        if (proc_sleeve_dir) {
                create_proc_read_entry("eeprom", 0, proc_sleeve_dir, h3600_sleeve_proc_read_eeprom, 0);
                create_proc_read_entry("backpaq", 0, proc_sleeve_dir, h3600_sleeve_proc_read_backpaq, 0);
        }

        /* this thread calls the handlers, so that they are not called in interrupt context */
        kernel_thread ( h3600_sleeve_thread, NULL, 
                        CLONE_FS | CLONE_FILES | CLONE_SIGHAND );
        wake_up_interruptible( &h3600_sleeve_wait );

	if (!result)
	{	
                printk("init_module successful init major= %d irq=%d\n",
                       h3600_sleeve_major, IRQ_GPIO_BITSY_OPT_DET);

	}
	else
		printk(__FUNCTION__ ": error from irq %d\n", result);
	return 0;
}

void h3600_sleeve_cleanup_module(void)
{
	printk(__FUNCTION__ ":\n");

        free_irq(IRQ_GPIO_BITSY_OPT_DET, h3600_sleeve_interrupt);

        if (proc_sleeve_dir) {
                remove_proc_entry("eeprom", proc_sleeve_dir);
                remove_proc_entry("sleeve", proc_bus);
                remove_proc_entry("backpaq", proc_bus);
        }

#if 0
	unregister_chrdev(, MOD_NAME);
#endif
}

module_init(h3600_sleeve_init_module);
module_exit(h3600_sleeve_cleanup_module);

