/*
 *  drivers/mtd/nand.c
 *
 *  Copyright (C) 2000 Steven J. Hill (sjhill@cotw.com)
 *
 * $Id: nand.c,v 1.2 2000/10/13 22:54:09 mdeans Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Overview:
 *   This is the generic MTD driver for NAND flash devices. It is
 *   capable of talking to all popular NAND chips currently available.
 */

#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/nand_ids.h>
#include <asm/io.h>

#ifdef CONFIG_MTD_NAND_ECC
#include <linux/mtd/nand_ecc.h>
#endif

#if LINUX_VERSION_CODE < 0x020300
#define set_current_state(x) current->state = (x);
#endif

/*
 * Macros for low-level register control
 */
#define NAND_CTRL (*(volatile unsigned char *) \
			((struct nand_chip *) mtd->priv)->CTRL_ADDR)
#define nand_select() NAND_CTRL &= ~this->NCE;
#define nand_deselect() NAND_CTRL |= ~this->NCE;

/*
 * NAND low-level MTD interface functions
 */
static int nand_read (struct mtd_info *mtd, loff_t from, size_t len,
			size_t *retlen, u_char *buf);
static int nand_read_ecc (struct mtd_info *mtd, loff_t from, size_t len,
				size_t *retlen, u_char *buf, u_char *ecc_code);
static int nand_read_oob (struct mtd_info *mtd, loff_t from, size_t len,
				size_t *retlen, u_char *buf);
static int nand_write (struct mtd_info *mtd, loff_t to, size_t len,
			size_t *retlen, const u_char *buf);
static int nand_write_ecc (struct mtd_info *mtd, loff_t to, size_t len,
				size_t *retlen, const u_char *buf,
				u_char *ecc_code);
static int nand_write_oob (struct mtd_info *mtd, loff_t to, size_t len,
				size_t *retlen, const u_char *buf);
static int nand_erase (struct mtd_info *mtd, struct erase_info *instr);

/*
 * Send command to NAND device
 */
static void nand_command (struct mtd_info *mtd, unsigned command,
				int column, int page_addr)
{
	register struct nand_chip *this = mtd->priv;
	register unsigned long NAND_IO_ADDR = this->IO_ADDR;

	/* Begin command latch cycle */
	NAND_CTRL |= this->CLE;

	/*
	 * Write out the command to the device. When writing
	 * data to the device the sequence is more complicated.
	 */
	if (command != NAND_CMD_SEQIN)	
		writeb (command, NAND_IO_ADDR);
	else {
		if (mtd->oobblock == 256 && column >= 256) {
			column -= 256;
			writeb(NAND_CMD_RESET, NAND_IO_ADDR);
			writeb(NAND_CMD_READOOB, NAND_IO_ADDR);
			writeb(NAND_CMD_SEQIN, NAND_IO_ADDR);
		}
		else if (mtd->oobblock == 512 && column >= 256) {
			if (column < 512) {
				column -= 256;
				writeb(NAND_CMD_READ1, NAND_IO_ADDR);
				writeb(NAND_CMD_SEQIN, NAND_IO_ADDR);
			}
			else {
				column -= 512;
				writeb(NAND_CMD_READOOB, NAND_IO_ADDR);
				writeb(NAND_CMD_SEQIN, NAND_IO_ADDR);
			}
		}
		else {
			writeb(NAND_CMD_READ0, NAND_IO_ADDR);
			writeb(NAND_CMD_SEQIN, NAND_IO_ADDR);
		}
	}

	/* Set ALE and clear CLE to start address cycle */
	NAND_CTRL &= ~this->CLE;
	NAND_CTRL |= this->ALE;

	/* Serially input address */
	if (column != -1)
		writeb (column, NAND_IO_ADDR);
	if (page_addr != -1) {
		writeb ((unsigned char) (page_addr & 0xff), NAND_IO_ADDR);
		writeb ((unsigned char) ((page_addr >> 8) & 0xff), NAND_IO_ADDR);
		/* One more address cycle for higher density devices */
		if (mtd->size & 0x0c000000) {
			writeb ((unsigned char) ((page_addr >> 16) & 0x0f),
					NAND_IO_ADDR);
		}
	}

	/* Latch in address */
	NAND_CTRL &= ~this->ALE;

	/* Pause for 15us */
	udelay (15);
}

/*
 * NAND read
 */
static int nand_read (struct mtd_info *mtd, loff_t from, size_t len,
			size_t *retlen, u_char *buf)
{
#ifdef CONFIG_MTD_NAND_ECC
	struct nand_chip *this = mtd->priv;
	
	return nand_read_ecc (mtd, from, len, retlen, buf, this->ecc_code_buf);
#else
	return nand_read_ecc (mtd, from, len, retlen, buf, NULL);
#endif
}

/*
 * NAND read with ECC
 */
static int nand_read_ecc (struct mtd_info *mtd, loff_t from, size_t len,
				size_t *retlen, u_char *buf, u_char *ecc_code)
{
	int j, column, page, ecc_result, temp, state;
	struct nand_chip *this = mtd->priv;
#ifdef CONFIG_MTD_NAND_ECC
	u_char ecc_calc[3];
#endif

	/* First we calculate the starting page */
	page = from >> this->page_shift;

	/* Do not allow reads past end of device */
	if (from >= mtd->size || (from + len) >= mtd->size) {
		DEBUG (MTD_DEBUG_LEVEL3,
			"nand_read_ecc: Attempt read beyond end of device\n");
		*retlen = 0;
		return -EINVAL;
	}

	/*
	 * Now we calculate the column and initialize the state machine
	 * for use with devices with 512 byte wide pages.
	 */
	if ((from - (page * mtd->oobblock)) < 256) {
		column = from - (page * mtd->oobblock);
		state = 0;
	}
	else {
		column = from - (page * mtd->oobblock) - 256;
		state = 1;
	}

	/* Initialize return value */
	*retlen = 0;

	/* Select the NAND device */
	nand_select ();

	/* Loop through the pages */
	while (1) {
		/* Send the read command */
		if (!state)
			nand_command (mtd, NAND_CMD_READ0, 0x00, page);
		else
			nand_command (mtd, NAND_CMD_READ1, 0x00, page);

		/* Read in 256 bytes */
		for (j=0 ; j<256 ; j++)
			this->data_buf[j] = readb (this->IO_ADDR);

#ifdef CONFIG_MTD_NAND_ECC
		/* Read in the out-of-band data */
		if (!state && mtd->oobblock == 512) {
			nand_command (mtd, NAND_CMD_READOOB, 0x00, page);
			for (j=0 ; j<3 ; j++)
				ecc_code[j] = readb(this->IO_ADDR);
			nand_command (mtd, NAND_CMD_READ0, 0x00, page);
		}
		else if (mtd->oobblock == 256) {
			for (j=0 ; j<3 ; j++)
				ecc_code[j] = readb (this->IO_ADDR);
		}
		else {
			readb (this->IO_ADDR);
			readb (this->IO_ADDR);
			readb (this->IO_ADDR);
			for (j=0 ; j<3 ; j++)
				ecc_code[j] = readb (this->IO_ADDR);
		}

		/* Calculate the ECC and verify it */
		nand_calculate_ecc (&this->data_buf[0],
					&ecc_calc[0]);
		ecc_result = nand_correct_data (&this->data_buf[0],
					&ecc_code[0] , &ecc_calc[0]);
		if (ecc_result == -1) {
			DEBUG (MTD_DEBUG_LEVEL1,
				"nand_read_ecc: " \
				"Failed read, page 0x%08x\n", page);
			nand_deselect ();
			return -EIO;
		}
#endif

		/* Check to see if we have read enough data */
		if ((*retlen + 256) >= len) {
			if (*retlen == 0) {
				temp = column + len;
				for (j=column ; j<temp ; j++)
					buf[(*retlen)++] = this->data_buf[j];
			}
			else {
				temp = len - *retlen;
				for (j=0 ; j<temp ; j++)
					buf[(*retlen)++] = this->data_buf[j];
			}
			return 0;
		}

		/*
		 * We either got a partial 256 byte block
		 * or an entire 256 byte block.
		 */
		if (*retlen == 0) {
			for (j=column ; j<256 ; j++)
				buf[(*retlen)++] = this->data_buf[j];
		}
		else {
			for (j=0 ; j<256 ; j++)
				buf[(*retlen)++] = this->data_buf[j];
		}

		/*
		 * Increment page address if we are a device with 256 byte
		 * wide pages or toggle the state machine for 512 byte
		 * devices and possibly increment the page address.
		 */
		if (mtd->oobblock == 256) {
			page++;
		}
		else if (!state) {
			state = 1;
		}
		else {
			state = 0;
			page++;
		}
	}

	/* De-select the NAND device */
	nand_deselect ();

	/* Return happy */
	return 0;
}

/*
 * NAND read out-of-band
 */
static int nand_read_oob (struct mtd_info *mtd, loff_t from, size_t len,
				size_t *retlen, u_char *buf)
{
	int i, column, page;
	struct nand_chip *this = mtd->priv;
	
	/* Shift to get page */
	page = ((int) from) >> this->page_shift;

	/* Mask to get column */
	column = from & 0x1f;

	/* Initialize return length value */
	*retlen = 0;

	/* Do not allow read past end of page */
	if ((column + len) > mtd->oobsize) {
		DEBUG (MTD_DEBUG_LEVEL3,
			"nand_read_oob: Attempt to read past end of page\n");
		return -EINVAL;
	}

	/* Select the NAND device */
	nand_select ();

	/* Send the read command */
	nand_command (mtd, NAND_CMD_READOOB, column, page);	

	/* Read the data */
	for (i = 0 ; i < len ; i++)
		buf[i] = readb (this->IO_ADDR);

	/* De-select the NAND device */
	nand_deselect ();

	/* Return happy */
	*retlen = len;
	return 0;
}

/*
 * NAND write
 */
static int nand_write (struct mtd_info *mtd, loff_t to, size_t len,
			size_t *retlen, const u_char *buf)
{
#ifdef CONFIG_MTD_NAND_ECC
	struct nand_chip *this = mtd->priv;
	
	return nand_write_ecc (mtd, to, len, retlen, buf, this->ecc_code_buf);
#else
	return nand_write_ecc (mtd, to, len, retlen, buf, NULL);
#endif
}

/*
 * NAND write with ECC
 */
static int nand_write_ecc (struct mtd_info *mtd, loff_t to, size_t len,
				size_t *retlen, const u_char *buf,
				u_char *ecc_code)
{
	int i, page, status;
	struct nand_chip *this = mtd->priv;
#ifdef CONFIG_MTD_NAND_ECC
	u_char ecc_calc[6];
#endif

	/* Do not allow unaligned writes */
	if ((to & (mtd->oobblock-1)) || (len & (mtd->oobblock-1))) {
		DEBUG (MTD_DEBUG_LEVEL1,
			"nand_write_ecc: Unaligned address or length\n");
		return -EINVAL;
	}

	/* Do not allow write past end of page */
	if ((to + len) > mtd->size) {
		DEBUG (MTD_DEBUG_LEVEL1,
			"nand_write_ecc: Attempted write past end of device\n");
		return -EINVAL;
	}

	/* Shift to get page */
	page = ((int) to) >> this->page_shift;

	/* Initialize return length value */
	*retlen = 0;

	/* Select the NAND device */
	nand_select ();

	/* Check the WP bit */
	nand_command (mtd, NAND_CMD_STATUS, -1, -1);
	if (!(readb (this->IO_ADDR) & 0x80)) {
		DEBUG (MTD_DEBUG_LEVEL0,
			"nand_write_ecc: Device is write protected!!!\n");
		nand_deselect ();
		return -EIO;
	}

	/* Loop through the pages */
	while (1) {
		/* Send command to begin auto page programming */
		nand_command (mtd, NAND_CMD_SEQIN, 0, page);

		/* Write out complete page of data */
		for (i=0 ; i < mtd->oobblock ; i++)
			writeb (buf[(*retlen + i)], this->IO_ADDR);

#ifdef CONFIG_MTD_NAND_ECC
		/* Calculate the ECC and write it */
		nand_calculate_ecc (&buf[*retlen], &ecc_calc[0]);
		for (i=0 ; i<3 ; i++)
			writeb (ecc_calc[i], this->IO_ADDR);

		/* Calculate second ECC for 512 byte page device */
		if (mtd->oobblock == 512) {
			nand_calculate_ecc (&buf[(*retlen + 256)], &ecc_calc[3]);
			for (i=3 ; i<6 ; i++)
				writeb (ecc_calc[i], this->IO_ADDR);
		}

		/* Write ones for partial page programming */
		if (mtd->oobblock == 512)
			for (i=0 ; i<10 ; i++)
				writeb (0xff, this->IO_ADDR);
		else
			for (i=0 ; i<5 ; i++)
				writeb (0xff, this->IO_ADDR);
#else
		/* Write ones for partial page programming */
		for (i=0 ; i < mtd->eccsize ; i++)
			writeb (0xff, this->IO_ADDR);
#endif

		/* Send command to actually program the data */
		nand_command (mtd, NAND_CMD_PAGEPROG, -1, -1);

		/*
		 * Wait for program operation to complete. This could
		 * take up to 3000us (3ms) on some devices, so we try
		 * and exit as quickly as possible.
		 */
		status = 0;
		for (i=0 ; i<24 ; i++) {
			/* Delay for 125us */
			udelay (125);

			/* Check the status */
			nand_command (mtd, NAND_CMD_STATUS, -1, -1);
			status = (int) readb (this->IO_ADDR);
			if (status & 0x40)
				break;
		}

		/* See if device thinks it succeeded */
		if (status & 0x01) {
			DEBUG (MTD_DEBUG_LEVEL1,
				"nand_write_ecc: " \
				"Failed write, page 0x%08x\n", page);
			nand_deselect ();
			return -EIO;
		}

#ifdef CONFIG_MTD_NAND_VERIFY_WRITE
		/*
		 * The NAND device assumes that it is always writing to
		 * a cleanly erased page. Hence, it performs its internal
		 * write verification only on bits that transitioned from
		 * 1 to 0. The device does NOT verify the whole page on a
		 * byte by byte basis. It is possible that the page was
		 * not completely erased or the page is becoming unusable
		 * due to wear. The read with ECC will catch this, but we
		 * would rather catch it early in the write stage. Better
		 * to write no data than invalid data.
		 */
		
		/* Send command to read back the page */
		nand_command (mtd, NAND_CMD_READ0, 0, page);

		/* Loop through and verify the data */
		for (i=0 ; i < mtd->oobblock ; i++) {
			if (buf[(*retlen + i)] != readb (this->IO_ADDR)) {
				DEBUG (MTD_DEBUG_LEVEL0,
					"nand_write_ecc: " \
					"Failed write verify, page 0x%08x\n", page);
				nand_deselect ();
				return -EIO;
			}
		}
#endif

		/* Update write count */
		*retlen += mtd->oobblock;

		/* All data written or increment page address */
		if ((*retlen - len) != 0)
			page++;
		else
			break;
	}

	/* De-select the NAND device */
	nand_deselect ();

	/* Return happy */
	*retlen = len;
	return 0;
}

/*
 * NAND write out-of-band
 */
static int nand_write_oob (struct mtd_info *mtd, loff_t to, size_t len,
				size_t *retlen, const u_char *buf)
{
	int i, column, page, status;
	struct nand_chip *this = mtd->priv;
	
	/* Shift to get page */
	page = ((int) to) >> this->page_shift;

	/* Mask to get column */
	column = to & 0x1f;

	/* Initialize return length value */
	*retlen = 0;

	/* Do not allow write past end of page */
	if ((column + len) > mtd->oobsize) {
		DEBUG (MTD_DEBUG_LEVEL1,
			"nand_write_oob: Attempt to write past end of page\n");
		return -EINVAL;
	}

	/* Select the NAND device */
	nand_select ();

	/* Check the WP bit */
	nand_command (mtd, NAND_CMD_STATUS, -1, -1);
	if (!(readb (this->IO_ADDR) & 0x80)) {
		DEBUG (MTD_DEBUG_LEVEL0,
			"nand_write_oob: Device is write protected!!!\n");
		nand_deselect ();
		return -EIO;
	}

	/* Write out desired data */
	nand_command (mtd, NAND_CMD_SEQIN, column + 512, page);
	for (i=0 ; i<len ; i++)
		writeb (buf[i], this->IO_ADDR);

	/* Send command to program the OOB data */
	nand_command (mtd, NAND_CMD_PAGEPROG, -1, -1);

	/*
	 * Wait for program operation to complete. This could
	 * take up to 3000us (3ms) on some devices, so we try
	 * and exit as quickly as possible.
	 */
	status = 0;
	for (i=0 ; i<24 ; i++) {
		/* Delay for 125us */
		udelay (125);

		/* Check the status */
		nand_command (mtd, NAND_CMD_STATUS, -1, -1);
		status = (int) readb (this->IO_ADDR);
		if (status & 0x40)
			break;
	}

	/* See if device thinks it succeeded */
	if (status & 0x01) {
		DEBUG (MTD_DEBUG_LEVEL1,
			"nand_write_oob: " \
			"Failed write, page 0x%08x\n", page);
		nand_deselect ();
		return -EIO;
	}

#ifdef CONFIG_MTD_NAND_VERIFY_WRITE
	/* Send command to read back the data */
	nand_command (mtd, NAND_CMD_READOOB, column, page);

	/* Loop through and verify the data */
	for (i=0 ; i<len ; i++) {
		if (buf[i] != readb (this->IO_ADDR)) {
			DEBUG (MTD_DEBUG_LEVEL1,
				"nand_write_oob: " \
				"Failed write verify, page 0x%08x\n", page);
			nand_deselect ();
			return -EIO;
		}
	}
#endif

	/* De-select the NAND device */
	nand_deselect ();

	/* Return happy */
	*retlen = len;
	return 0;
}

/*
 * NAND erase a block
 */
static int nand_erase (struct mtd_info *mtd, struct erase_info *instr)
{
	int i, page, len, status, pages_per_block, sleep_cnt;
	struct nand_chip *this = mtd->priv;

	/* Start address must align on block boundary */
	if (instr->addr & (mtd->erasesize - 1)) {
		DEBUG (MTD_DEBUG_LEVEL1,
			"nand_erase: Unaligned address\n");
		return -EINVAL;
	}

	/* Length must align on block boundary */
	if (instr->len & (mtd->erasesize - 1)) {
		DEBUG (MTD_DEBUG_LEVEL1,
			"nand_erase: Length not block aligned\n");
		return -EINVAL;
	}

	/* Do not allow erase past end of device */
	if ((instr->len + instr->addr) > mtd->size) {
		DEBUG (MTD_DEBUG_LEVEL1,
			"nand_erase: Erase past end of device\n");
		return -EINVAL;
	}

	/* Shift to get first page */
	page = (int) (instr->addr >> this->page_shift);

	/* Calculate pages in each block */
	pages_per_block = mtd->erasesize / mtd->oobblock;

	/* Select the NAND device */
	nand_select ();

	/* Check the WP bit */
	nand_command (mtd, NAND_CMD_STATUS, -1, -1);
	if (!(readb (this->IO_ADDR) & 0x80)) {
		DEBUG (MTD_DEBUG_LEVEL0,
			"nand_erase: Device is write protected!!!\n");
		nand_deselect ();
		return -EIO;
	}

	/* Loop through the pages */
	len = instr->len;
	sleep_cnt = 0;
	while (len) {
		/* Send commands to erase a page */
		nand_command(mtd, NAND_CMD_ERASE1, -1, page);
		nand_command(mtd, NAND_CMD_ERASE2, -1, -1);

		/*
		 * Wait for program operation to complete. This could
		 * take up to 4000us (4ms) on some devices, so we try
		 * and exit as quickly as possible.
		 */
		status = 0;
		for (i=0 ; i<32 ; i++) {
			/* Delay for 125us */
			udelay (125);

			/* Check the status */
			nand_command (mtd, NAND_CMD_STATUS, -1, -1);
			status = (int) readb (this->IO_ADDR);
			if (status & 0x40)
				break;
		}

		/* See if block erase succeeded */
		if (status & 0x01) {
			DEBUG (MTD_DEBUG_LEVEL1,
				"nand_erase: " \
				"Failed erase, page 0x%08x\n", page);
			nand_deselect ();
			return -EIO;
		}

		/* Increment page address and decrement length */
		len -= mtd->erasesize;
		page += pages_per_block;

		/*
		 * If we are erasing more than 2 blocks, don't let
		 * this operation eat an entire clock period.
		 */
/*
		sleep_cnt++;
		if ((sleep_cnt >= 2) && (sleep_cnt % 2)) {
			current->timeout = jiffies + 1;
			current->state = TASK_INTERRUPTIBLE;
			schedule ();
			current->timeout = 0;
		}
*/
	}

	/* De-select the NAND device */
	nand_deselect ();

	/* Return happy */
	return 0;
}

/*
 * Scan for the NAND device
 */
int nand_scan (struct mtd_info *mtd)
{
	int i, nand_maf_id, nand_dev_id;
	struct nand_chip *this = mtd->priv;
	struct erase_info *sjh_erase;
	int count;
	u_char *buf;

	/* Select the device */
	nand_select ();

	/* Send the command for reading device ID */
	nand_command (mtd, NAND_CMD_READID, 0x00, -1);

	/* Read manufacturer and device IDs */
	nand_maf_id = readb (this->IO_ADDR);
	nand_dev_id = readb (this->IO_ADDR);

	/* Print and store flash device information */
	for (i = 0; nand_flash_ids[i].name != NULL; i++) {
		if (nand_maf_id == nand_flash_ids[i].manufacture_id &&
		    nand_dev_id == nand_flash_ids[i].model_id) {
			if (!mtd->size) {
				mtd->name = nand_flash_ids[i].name;
				mtd->erasesize = nand_flash_ids[i].erasesize;
				mtd->size = (1 << nand_flash_ids[i].chipshift);
				mtd->eccsize = 256;
				if (nand_flash_ids[i].page256) {
					mtd->oobblock = 256;
					mtd->oobsize = 8;
					this->page_shift = 8;
				}
				else {
					mtd->oobblock = 512;
					mtd->oobsize = 16;
					this->page_shift = 9;
				}
			}
			printk (KERN_INFO "NAND device: Manufacture ID:" \
				" 0x%02x, Chip ID: 0x%02x (%s)\n",
			       nand_maf_id, nand_dev_id, mtd->name);
			break;
		}
	}

	/* Initialize state and spinlock */
	this->state = FL_READY;
	this->chip_lock = SPIN_LOCK_UNLOCKED;

	/* De-select the device */
	nand_deselect ();

	/* Print warning message for no device */
	if (!mtd->size) {
		printk (KERN_WARNING "No NAND device found!!!\n");
		return 1;
	}

	/* Fill in remaining MTD driver data */
	mtd->type = MTD_NANDFLASH;
	mtd->flags = MTD_CAP_NANDFLASH | MTD_ECC;
	mtd->module = THIS_MODULE;
	mtd->erase = nand_erase;
	mtd->point = NULL;
	mtd->unpoint = NULL;
	mtd->read = nand_read;
	mtd->write = nand_write;
	mtd->ecctype = MTD_ECC_SW;
	mtd->read_ecc = nand_read_ecc;
	mtd->write_ecc = nand_write_ecc;
	mtd->read_oob = nand_read_oob;
	mtd->write_oob = nand_write_oob;
	mtd->sync = NULL;

	/* Return happy */
	return 0;
}
